import os
import yaml
from pathlib import Path
from git import Repo
import tempfile
from typing import Generator, Callable

from .base import BaseTemplate, TemplateProperties


class GitTemplate(BaseTemplate):
    template_properties_filename = 'template_properties.yaml'
    variables_helper_filename = 'variables_helper.py'

    def __init__(self, template_name: str, git_repo_path: str):
        self.template_name = template_name
        self._documents = {}
        
        # Use TemporaryDirectory as a context manager
        with tempfile.TemporaryDirectory() as temp_dir:
            Repo.clone_from(git_repo_path, temp_dir)
            self.properties = self._load_properties(temp_dir)
            self.variables_helper: Callable[[dict], dict] = self._load_variables_helper(temp_dir)
            self._load_documents(temp_dir)

    def _load_properties(self, temp_dir: str) -> TemplateProperties:
        properties_filename = Path(temp_dir) / self.template_properties_filename
        if not os.path.exists(properties_filename):
            raise FileNotFoundError(f"Template properties file '{properties_filename}' does not exist.")
        with open(properties_filename, 'r') as file:
            properties = yaml.safe_load(file)
        return properties or {}

    def _load_variables_helper(self, temp_dir) -> Callable:
        variables_helper_filename = Path(temp_dir) / Path(self.variables_helper_filename)
        if not os.path.exists(variables_helper_filename):
            return lambda d: d
        with open(variables_helper_filename, 'r') as file:
            code = file.read()
        exec_globals = {}
        exec(code, exec_globals)
        variables_helper = exec_globals.get('variables_helper')
        if not callable(variables_helper):
            raise ValueError(f"Variables helper in '{variables_helper_filename}' is not callable.")
        return variables_helper

    def _load_documents(self, temp_dir: str):
        template_root = Path(temp_dir) / "template"
        if not template_root.exists():
            raise FileNotFoundError(f"Template root directory '{template_root}' does not exist.")
        for root, dirs, files in os.walk(template_root):
            rel_root = Path(root).relative_to(template_root)
            for name in files:
                rel_path_template = rel_root / name
                abs_path_template = template_root / rel_path_template
                with open(abs_path_template, 'r') as file:
                    content = file.read()
                self._documents[str(rel_path_template)] = content

    def documents(self) -> Generator[tuple[str, str], None, None]:
        """
        Yields stored (relpath, content) tuples from the document dictionary.
        """
        for relpath, content in self._documents.items():
            yield relpath, content



from typing import Any, TypedDict, Literal


class CustomVariable(TypedDict):
    name: str
    type: str | None
    default: Any | None
    description: str | None


class TemplateProperties(TypedDict):
    custom_variables: list[CustomVariable] | None
    templater: Literal["pystring", "jinja2"] | None
    auto_use_defaults: bool | None



import pytest
from unittest.mock import patch, MagicMock

from skaf.templaters.base import ABCTemplater


class TestBaseTemplater:
    
    def test_abc_templater_cannot_be_instantiated(self):
        # Verify that ABCTemplater cannot be instantiated directly
        with pytest.raises(TypeError):
            ABCTemplater()
    
    def test_abc_templater_requires_render_method(self):
        # Create a subclass that doesn't implement render
        class IncompleteTemplater(ABCTemplater):
            pass
        
        # Verify that instantiating the subclass raises TypeError
        with pytest.raises(TypeError):
            IncompleteTemplater()
    
    def test_abc_templater_with_render_implementation(self):
        # Create a concrete implementation
        class ConcreteTemplater(ABCTemplater):
            def render(self, template, context):
                return f"Rendered: {template} with {context}"
        
        # Instantiate and verify it works
        templater = ConcreteTemplater()
        result = templater.render("test template", {"var": "value"})
        
        assert result == "Rendered: test template with {'var': 'value'}"



import pytest
from unittest.mock import patch, MagicMock

from skaf.template_classes.dict_template import DictTemplate
from skaf.template_classes.base import TemplateProperties


class TestDictTemplate:
    
    def test_dict_template_init(self):
        # Setup test data
        template_name = "test_template"
        properties = {"name": "Test Template", "description": "A test template"}
        templates = {
            "file1.py": "content1",
            "file2.py": "content2"
        }
        
        # Create instance
        template = DictTemplate(template_name, properties, templates)
        
        # Verify attributes
        assert template.template_name == template_name
        assert template.properties == properties
        assert hasattr(template, "templates")
        assert template.templates == templates
    
    def test_dict_template_documents(self):
        # Setup test data
        templates = {
            "file1.py": "content1",
            "file2.py": "content2",
            "file3.py": "content3"
        }
        
        # Create instance
        template = DictTemplate("test", {}, templates)
        
        # Get documents
        documents = list(template.documents())
        
        # Verify results
        assert len(documents) == 3
        assert ("file1.py", "content1") in documents
        assert ("file2.py", "content2") in documents
        assert ("file3.py", "content3") in documents



from string import Template

from .base import ABCTemplater


class PystringTemplater(ABCTemplater):

    suffix = ".template"

    def render(self, template: str, context: dict, template_filename: str = None) -> str:
        """
        Render a template with the given context using Python string templating.

        Args:
            template (str): The template to render.
            context (dict): The context to use for rendering.

        Returns:
            str: The rendered template.
        """
        if template_filename:
            if not template_filename.endswith(self.suffix):
                return template
        pystring_template = Template(template)
        return pystring_template.safe_substitute(context)



import re


def sanitize_project_name(name: str) -> str:
    """
    Converts the provided project name to a valid Python package name.
    Lowercases the name, replaces non-alphanumeric characters with underscores,
    and prepends an underscore if the name starts with a digit.
    """
    name = name.lower()
    sanitized = re.sub(r'\W+', '_', name)
    if sanitized and sanitized[0].isdigit():
        sanitized = "_" + sanitized
    return sanitized



import os
import sys
from pathlib import Path
import yaml
from typing import Any

from ..template_classes.base import BaseTemplate
from ..templaters.base import ABCTemplater
from .variables import get_variable_values
from .context import ScaffoldContext


template_lib_dir = Path(__file__).parent / 'template_lib'
TEMPLATE_PROPERTIES_FILENAME = 'template_properties.yaml'


def apply_templating(document: str,
                     variables: dict[str, Any],
                     templater: ABCTemplater,
                     document_filename: str = None
                     ) -> str:
    """
    Applies templating to the given document string using the provided variables
    using the specified templater.
    """
    try:
        return templater.render(document, variables, template_filename=document_filename)
    except KeyError as e:
        raise ValueError(f"Missing variable for templating: {e}")
    except Exception as e:
        raise RuntimeError(f"Error applying templating: {e}")


def get_package_template_dir(template_name: str) -> Path:
    """
    Returns the path to the specified template directory.
    If the template does not exist, raises a FileNotFoundError.
    """
    template_path = template_lib_dir / template_name
    if not template_path.exists():
        raise FileNotFoundError(f"Template '{template_name}' does not exist.")
    return template_path


def load_template_properties(template_name) -> dict:
    template_dir = get_package_template_dir(template_name)
    properties_filename = template_dir / TEMPLATE_PROPERTIES_FILENAME
    with open(properties_filename, 'r') as file:
        properties = yaml.safe_load(file)
    return properties


def get_template_variable_values(context: ScaffoldContext) -> dict[str, Any]:
    try:
        return get_variable_values(context)
    except Exception as e:
        if context._debug:
            raise
        etype = type(e).__name__
        print(f"Error getting variable values: {etype}: {e}")
        sys.exit(1)


def map_paths(context: ScaffoldContext,
              variables: dict[str, Any]
              ) -> dict[Path, str]:
    """
    Using the `template.documents` iterator method, get the `relpath, content` pairs
    and create a new `target_path, content` mapping. Perform templating on the template
    relpath using the provided variables.
    """
    targets = {}
    for relpath, content in context.template.documents():
        relpath = Path(apply_templating(relpath, variables, context.templater))
        target_path = context.project_path / relpath
        target_path.parent.mkdir(parents=True, exist_ok=True)
        targets[relpath] = content
    return targets


def scaffold_project(project_name: str,
                     template_name: str = None,
                     output_dir: str = None,
                     overwrite: bool = False,
                     template: BaseTemplate = None,
                     auto_use_defaults: bool = True,
                     varfile: str | None = None,
                     no_project_dir: bool = False,
                     _debug: bool = False
                     ) -> None:
    """
    Scaffold a new project based on the provided template and variables.
    Copies files from the template directory to the new project directory,
    replacing placeholders with the provided variable values.
    """
    if output_dir is None:
        output_dir = os.getcwd()
    output_dir = Path(output_dir)

    context = ScaffoldContext(
        project_name=project_name,
        template_name=template_name,
        output_dir=output_dir,
        no_project_dir=no_project_dir,
        overwrite=overwrite,
        auto_use_defaults=auto_use_defaults,
        template=template,
        variables_filepath=Path(varfile) if varfile else None,
        _debug=_debug
    )

    variables = get_template_variable_values(context)

    path_mapping: dict[Path, str] = map_paths(
        context,
        variables
    )

    if not context.overwrite:
        if all([
            context.project_path.exists(),
            context.project_path.is_dir(),
            os.listdir(context.project_path)
        ]):
            print(f"Project directory '{context.project_path}' already exists. Set --overwrite to overwrite.")
            sys.exit(1)

    context.project_path.mkdir(parents=True, exist_ok=True)

    for target_path, content in path_mapping.items():
        target_filename = target_path.name
        content = apply_templating(
            content,
            variables,
            context.templater,
            target_filename
        )
        if target_path.suffix == context.templater.suffix:
            target_path = Path(target_path.parent / target_path.stem)
        write_path = context.project_path / target_path
        write_path.parent.mkdir(parents=True, exist_ok=True)
        with open(write_path, 'w') as file:
            file.write(content)



from typing import Any
from .context import ScaffoldContext
import re
import os
import yaml
from pathlib import Path

ENV_VAR_PREFIX = "SKAF_"


custom_var_type_mapper = {
    'str': str,
    'int': int,
    'float': float,
    'bool': bool,
    'list': lambda x: [x.strip() for x in x.split(',')],
    'dict': lambda x: dict(item.split('=') for item in x.split(',')),
}


def add_project_name_variables(project_name: str, variables):
    """
    Adds project name and several other derivatives (project_name_snake, project_name_pascal, project_name_kebab)"""
    variables['project_name'] = project_name
    project_name_snake = re.sub(r'(?<!^)(?=[A-Z])', '_', project_name).lower().replace("-", "_")
    variables['project_name_snake'] = project_name_snake
    variables['project_name_pascal'] = ''.join(word.capitalize() for word in project_name_snake.split('_'))
    variables['project_name_kebab'] = project_name_snake.replace('_', '-')
    variables['project_name_title'] = project_name.replace("_", " ").title()
    return variables


def get_env_variable(name: str) -> Any:
    """
    Get the value of an environment variable, or None if not set.
    """
    env_var = ENV_VAR_PREFIX + str(name)
    return os.environ.get(env_var)


def load_variables_filepath(filepath: Path) -> dict[str, Any]:
    """
    Load variables from a YAML file.
    """
    if not filepath.exists():
        raise FileNotFoundError(f"Variables file '{filepath}' does not exist.")
    with open(filepath, 'r') as file:
        variables = yaml.safe_load(file)
    return variables


def get_variable_values(context: ScaffoldContext) -> dict[str, Any]:
    values = {}

    values_from_file = {}
    if context.variables_filepath:
        values_from_file = load_variables_filepath(context.variables_filepath)

    for custom_var in context.template.custom_variables:
        varname = custom_var['name']
        vartype = custom_var.get('type', 'str')
        caster = custom_var_type_mapper.get(vartype, str)
        default = custom_var.get('default')
        if (from_env := get_env_variable(varname)) is not None:
            try:
                values[varname] = caster(from_env)
                continue
            except Exception as e:
                raise type(e)(f"Environment variable {from_env} cannot be used with caster {caster}: {e}")
        if varname in values_from_file:
            try:
                values[varname] = caster(values_from_file[varname])
                continue
            except Exception as e:
                raise type(e)(f"Variable {varname} cannot be used with caster {caster}: {e}")
        if context.auto_use_defaults and default is not None:
            try:
                values[varname] = caster(default)
            except Exception as e:
                raise type(e)(f"Default value for {varname}, {default} cannot be used with caster {caster}: {e}")
        else:
            defaultstr = f" [{default}]" if default else ""
            val = input(f"Enter value for {varname} ({vartype}){defaultstr}: ")
            if not val and default is not None:
                val = default
            try:
                values[varname] = caster(val)
            except Exception as e:
                raise type(e)(f"Invalid value for {varname} with type {vartype} and caster {caster}: {e}")
                
    values = add_project_name_variables(context.project_name, values)
    if context.template.variables_helper:
        values = context.template.variables_helper(values)
    return values



import sys
import os
from pathlib import Path
from .scaffold import scaffold_project
from argparse import ArgumentParser
from .template_classes.filesystem_template import FilesystemTemplate
from .template_classes.git_template import GitTemplate


def get_args():
    parser = ArgumentParser(description="Run the templater to build out a project file structure from templates.")
    parser.add_argument("name", help="The name of the project to create.")
    parser.add_argument("-t", "--template", default=None, help="Name of the project template to use.")
    parser.add_argument("-p", "--path", default=None, help="Path to a template directory.")
    parser.add_argument("--varfile", default=None, help="Path to a yaml file holding variables values.")
    parser.add_argument("-g", "--git", default=None, help="URI of a git repo to be used as a template directory.")
    parser.add_argument("-o", "--output", help="Output directory for the project.", default=os.getcwd())
    parser.add_argument("--overwrite", action="store_true", help="Force overwrite existing files.")
    parser.add_argument("--auto-use-defaults", action="store_true", help="Automatically use default values for template variables if present. (Overrides the template properties field of the same name.)")
    parser.add_argument("--debug", action="store_true", help="Enable debug mode.")
    parser.add_argument("--no-project-dir", action="store_true", help="Do not create a project directory.")
    args = parser.parse_args()
    if args.auto_use_defaults is False:
        args.auto_use_defaults = None  # tracks only explicit True
    return args


def get_filesystem_template(template_path) -> FilesystemTemplate:
    """
    Get a template from the filesystem.
    """
    if os.path.isdir(template_path):
        template_name = Path(template_path).name
        template = FilesystemTemplate(template_name, template_path)
        return template
    else:
        raise ValueError(f"Template path '{template_path}' is not a valid directory.")


def get_git_template(git_uri) -> GitTemplate:
    """
    Get a template from a git repository.
    """
    template_name = Path(git_uri).name
    return GitTemplate(template_name, git_uri)


def main():
    args = get_args()
    project_name = args.name
    template_name = args.template
    output_dir = args.output
    template_path = args.path

    template = None
    if template_path:
        template = get_filesystem_template(template_path)
        template_name = template.template_name
    elif args.git:
        template = get_git_template(args.git)
        template_name = template.template_name

    try:
        scaffold_project(
            project_name=project_name,
            template_name=template_name,
            output_dir=output_dir,
            no_project_dir=args.no_project_dir,
            overwrite=args.overwrite,
            template=template,
            auto_use_defaults=args.auto_use_defaults,
            varfile=args.varfile,
            _debug=args.debug
            )
        print(f"Project '{project_name}' initialized successfully using the '{template_name}' template.")
    except Exception as e:
        if args.debug:
            raise
        etype = type(e).__name__
        print(f"An error occurred while initializing the project: {etype}: {e}")
        sys.exit(1)
